"""Module to manage time related calculations."""

import os
from datetime import datetime, timedelta
from pytz import timezone


DEFAULT_TIMEDELTA = 15 * 60  # 15 minutes
default_timezone_str = os.getenv("TZ", "Europe/Madrid")


def current_timestamp() -> int:
    """Return current timestamp."""
    return int(current_datetime().timestamp())
    # return int(time.time())


# def current_timestamp_with_timedelta(delta_in_seconds: int = DEFAULT_TIMEDELTA) -> int:
#     """Return current timestamp plus a timedelta."""
#     return current_timestamp() + delta_in_seconds


def current_datetime(time_zone_str=default_timezone_str) -> datetime:
    """Generate current datetime for given timezone."""
    time_zone = timezone(time_zone_str)
    return datetime.now(time_zone)


def datetime_from_timestamp(timestamp, time_zone_str=default_timezone_str) -> datetime:
    """Return datetime for the given timestamp."""
    time_zone = timezone(time_zone_str)
    return datetime.fromtimestamp(timestamp, tz=time_zone)


def current_datetime_with_timedelta(
    time_zone=default_timezone_str,
    days: float = 0,
    seconds: float = 0,
    microseconds: float = 0,
    milliseconds: float = 0,
    minutes: float = 0,
    hours: float = 0,
    weeks: float = 0,
) -> datetime:
    """Get current datetime plus a timedelta."""
    ret_datetime = current_datetime(time_zone) + timedelta(
        days=days,
        seconds=seconds,
        microseconds=microseconds,
        milliseconds=milliseconds,
        minutes=minutes,
        hours=hours,
        weeks=weeks,
    )
    return ret_datetime


def current_timestamp_with_timedelta(
    time_zone=default_timezone_str,
    days: float = 0,
    seconds: float = 0,
    microseconds: float = 0,
    milliseconds: float = 0,
    minutes: float = 0,
    hours: float = 0,
    weeks: float = 0,
) -> int:
    """Get current timestamp plus a timedelta."""
    ret_datetime = current_datetime_with_timedelta(
        time_zone=time_zone,
        days=days,
        seconds=seconds,
        microseconds=microseconds,
        milliseconds=milliseconds,
        minutes=minutes,
        hours=hours,
        weeks=weeks,
    )
    return int(ret_datetime.timestamp())



# -*- coding: utf-8 -*-
"""CRUD helpers for SQLAlchemy."""

from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
import logging


logger = logging.getLogger(__name__)


# Generic CRUD functions ##########################################################################
# READ
async def get_list(
    db: AsyncSession, model, offset: int = None, limit: int = None, order_by: str = None
):
    """Retrieve a list of elements from database"""
    stmt = select(model)
    stmt = add_options_to_statement(stmt, offset, limit, order_by, model)
    return await get_list_statement_result(db, stmt)


async def get_element_by_id(db: AsyncSession, model, element_id):
    """Retrieve any DB element by id."""
    return await db.get(model, element_id)


async def get_list_statement_result(db: AsyncSession, stmt):
    """Execute given statement and return list of items."""
    result = await db.execute(stmt)
    item_list = result.unique().scalars().all()
    return item_list


async def get_first_statement_result(db: AsyncSession, stmt):
    """Execute given statement and return the first item."""
    result = await db.execute(stmt)
    item = result.scalar()
    return item


# DELETE
async def delete_element_by_id(db: AsyncSession, model, element_id):
    """Delete any DB element by id."""
    element = await get_element_by_id(db, model, element_id)
    if element is not None:
        await db.delete(element)
        await db.commit()
    return element


# CREATE
async def create_element(db: AsyncSession, db_element):
    """Insert a element in the database."""

    try:
        db.add(db_element)
        await db.commit()
        await db.refresh(db_element)
    except Exception as exc:
        logger.error(exc)
        await db.rollback()
        db_element = None
    return db_element


# Optional parameters #############################################################################
def set_order_by_to_statement(stmt, model, order_by):
    """Adds order by to given statement if needed."""
    if order_by is not None:
        if order_by.startswith("-"):
            stmt = stmt.order_by(getattr(model, order_by[1:]).desc())
        else:
            stmt = stmt.order_by(getattr(model, order_by))
    return stmt


def add_options_to_statement(stmt, offset, limit, order_by, model):
    """Adds options to given statement if needed: offset, limit, order_by."""
    stmt = set_order_by_to_statement(stmt, model, order_by)
    if offset is not None:
        stmt = stmt.offset(offset)
    if limit is not None:
        stmt = stmt.limit(limit)
    return stmt



import os
import logging


logger = logging.getLogger(__name__)


def get_public_key_from_url(public_key_url: str):
    """Upload public key from URL."""

    import httpx

    public_key = None

    if public_key_url is None or public_key_url == "":
        logger.warning("No public key URL provided")
        return None

    try:
        with httpx.Client() as client:
            response = client.get(public_key_url)

            if response.status_code != 200:
                raise Exception(f"Status code: {response.status_code}")

            public_key = response.text

    except Exception as e:
        logger.warning("Could not load public key from URL: " + str(e))

    return public_key


def get_public_key_from_file(file_path: str):
    """Get public key from pem file."""
    public_key = None
    try:
        if not os.path.isfile(file_path):
            raise Exception("Public key file not found")

        with open(file_path, "r") as f:
            public_key = f.read()

    except Exception as e:
        logger.warning("Could not load public key from file: " + str(e))

    return public_key


def write_public_key_to_file(public_key: str, file_path: str):
    """Write public key to file."""
    try:
        with open(file_path, "w") as f:
            f.write(public_key)
    except Exception as e:
        logger.warning("Could not write public key to file: " + str(e))


def get_public_key(
    public_key: str = None,
    public_key_url: str = None,
    public_key_file_path: str = None,
):
    """Get public key."""

    if public_key is None or public_key == "":
        public_key = get_public_key_from_url(public_key_url)

    if public_key is None or public_key == "":
        public_key = get_public_key_from_file(public_key_file_path)

    return public_key


class Config:
    public_key = None
    public_key_url = os.getenv("PUBLIC_KEY_URL", "")
    jwt_issuer = os.getenv("JWT_ISSUER", "mu-sse")
    jwt_algorithm = os.getenv("JWT_ALGORITHM", "RS256")
    public_key_file_path = os.getenv("PUBLIC_KEY_FILE_PATH", "public_key.pem")

    def __init__(self):
        self.update_public_key()

    def update_public_key(self, public_key=None, public_key_file_path=None):
        """Update public key."""
        if public_key_file_path is not None:
            self.public_key_file_path = public_key_file_path

        public_key = get_public_key(
            public_key, self.public_key_url, self.public_key_file_path
        )

        if public_key is not None and public_key != "":
            self.public_key = public_key
            write_public_key_to_file(self.public_key, self.public_key_file_path)

        else:
            logger.error("No public key available")


config = Config()



import jwt
from mumichaspy.fastapi_jwt_chassis.config import config
from mumichaspy.fastapi_jwt_chassis.time import (
    current_timestamp,
    current_timestamp_with_timedelta,
    DEFAULT_TIMEDELTA,
)

TESTING_PUBLIC_KEY = """
-----BEGIN PUBLIC KEY-----
MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCUQJHeEDVKZAVKpYwXPxIGA4Uf
sV1Y15JlnoK3nr6bFUCdI3oKc/nEco5XY6Do4Ek+ECk2rnpxDExLAUJbA1ea41tt
jhGp7pJYwqU+eA66mu9lIkBjQFXagmvAAG+PfN4NE80DM17eLMcBXS2jkz2O36Iz
xtfFHeD8pDuYRtx3KwIDAQAB
-----END PUBLIC KEY-----
"""

TESTING_PRIVATE_KEY = """
-----BEGIN RSA PRIVATE KEY-----
MIICXAIBAAKBgQCUQJHeEDVKZAVKpYwXPxIGA4UfsV1Y15JlnoK3nr6bFUCdI3oK
c/nEco5XY6Do4Ek+ECk2rnpxDExLAUJbA1ea41ttjhGp7pJYwqU+eA66mu9lIkBj
QFXagmvAAG+PfN4NE80DM17eLMcBXS2jkz2O36IzxtfFHeD8pDuYRtx3KwIDAQAB
AoGAHOz4SP6alzgqvCTOz55Tsu6u478kengdLUyfLpp/kBy2bbKFVlLIkebUFQLw
W1+fisd2cx/Z7gK7RAsr2jwttFIZ2qaq3z+hMQ+ihqvMv5kQPMTiyymDCvNdrp3I
LTDz0yF4nygXvU3DR+vG1WQnzNWP4H5A7daBi3INoew+CVECQQDwypjbNKYhgMlY
r3yZw0nXrEJZM7x9i7+NUfmZsKn5olUmImm67rvMw7bbrCvxwgRFcCDo/G2iGo1N
MUpvrKsDAkEAnZ2wWfFMEHz+2b3k+jwTsSWrV3FaTBCWb0Wo8TDAsVHb+qUAUhTG
B/3kUimdXoz/1YqpnfvF2M1NiKEJokD2uQJACj1jYLcNdZFHfrj/wH44jq5pp+d2
VZRy0eyxg6Kqp84KdvOxsVxjqiKMjSz2NXK9wcIx9BaABwvABI9bDgK2xwJAO6dG
64Or6kGkovyAAUVjnpRHOR0ps8hDZ3HR8xcklvgIbeCtMmOyX6v0EjS13uc2nLeS
Yry7sMZjJebtIn+74QJBAOXUcGRQG6rJgWcI43mr40L1JanOPTuZvRHd2XfT9zkd
9ma0n1LI2LExC0U9Ho+3bMiHVK5lf7NhHgdqps3lMos=
-----END RSA PRIVATE KEY-----
"""

DECODED_MOCK_JWT = {
    "sub": 1,
    "username": "jdoe",
    "iss": config.jwt_issuer,
    "is_active": True,
    "email": "jdoe@example.com",
    "roles": ["USER"],
}

DECODED_ADMIN_MOCK_JWT = {
    "sub": 1,
    "username": "admin",
    "iss": config.jwt_issuer,
    "is_active": True,
    "email": "admin@example.com",
    "roles": ["ADMIN"],
}


def get_encoded_mock_jwt(jwt_payload):
    """Get an encoded JWT."""
    if "iat" not in jwt_payload:
        jwt_payload["iat"] = current_timestamp()
    if "exp" not in jwt_payload:
        jwt_payload["exp"] = current_timestamp_with_timedelta(seconds=DEFAULT_TIMEDELTA)
    return jwt.encode(jwt_payload, TESTING_PRIVATE_KEY, algorithm="RS256")


def mock_jwt_decode_error(encoded_token, public_key, issuer, algorithms):
    """Mock JWT decode error."""

    raise jwt.exceptions.DecodeError("error mocked")


def mock_jwt_decode_ok(encoded_token, public_key, issuer, algorithms):
    """Mock validate and decode token and return a JWT."""
    global DECODED_MOCK_JWT
    return {**DECODED_MOCK_JWT}


def mock_jwt_decode_admin_ok(encoded_token, public_key, issuer, algorithms):
    """Mock validate and decode token and return a JWT."""
    global DECODED_ADMIN_MOCK_JWT
    return {**DECODED_ADMIN_MOCK_JWT}


def mock_jwt_headers(headers: dict = {}):
    """Get a mock JWT header."""
    global DECODED_MOCK_JWT
    return {
        **headers,
        "Authorization": f"Bearer {get_encoded_mock_jwt(DECODED_MOCK_JWT)}",
    }


def mock_jwt_headers_admin(headers: dict = {}):
    """Get a mock JWT header."""
    global DECODED_ADMIN_MOCK_JWT
    return {
        **headers,
        "Authorization": f"Bearer {get_encoded_mock_jwt(DECODED_ADMIN_MOCK_JWT)}",
    }


def assert_token_validation_called(mock_validate_and_decode_token, encoded_token):
    """Assert mock validate and decode token was called."""
    mock_validate_and_decode_token.assert_called_once_with(
        encoded_token=encoded_token,
        public_key=config.public_key,
        issuer=config.jwt_issuer,
        algorithms=[config.jwt_algorithm],
    )



# -*- coding: utf-8 -*-
"""Dependencies for dependency injection."""

from sqlalchemy.orm import sessionmaker, declarative_base
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession
from mumichaspy.sqlalchemy_chassis.config import config

engine = create_async_engine(config.SQLALCHEMY_DATABASE_URL, echo=True)
SessionLocal = sessionmaker(
    autocommit=False, autoflush=False, bind=engine, class_=AsyncSession, future=True
)

Base = declarative_base()


async def get_db():
    """Get database."""

    db = SessionLocal()
    try:
        yield db
        await db.commit()
    except:  # noqa: E722
        await db.rollback()
    finally:
        await db.close()



import os
import uuid
import logging
from contextlib import asynccontextmanager
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession, AsyncEngine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import StaticPool
from mumichaspy.sqlalchemy_chassis.database import Base


logger = logging.getLogger(__name__)


async def reset_db(engine: AsyncEngine):
    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.drop_all)
        await conn.run_sync(Base.metadata.create_all)


@asynccontextmanager
async def get_unique_memory_db():
    """Get unique meomory engine and db (e.g. for testing)."""

    engine_filename = f"file:{uuid.uuid4().hex}"
    engine = create_async_engine(
        # f"sqlite+aiosqlite://:memory:?cache=shared&uuid={uuid.uuid4()}",
        f"sqlite+aiosqlite:///{engine_filename}?mode=memory&cache=shared",
        poolclass=StaticPool,
        echo=False,
    )

    TestingSessionLocal = sessionmaker(
        autocommit=False,
        autoflush=False,
        bind=engine,
        class_=AsyncSession,
        future=True,
    )
    db = TestingSessionLocal()
    await reset_db(engine)
    try:
        yield db
    except Exception as exc:
        raise exc
    finally:
        await db.close()
        logger.info(f"Closed db {db}")
        await engine.dispose()
        logger.info(f"Disposed engine {engine}")
        if os.path.exists(engine_filename):
            os.remove(engine_filename)
            print(f"Removed {engine_filename}")



"""Security module for JWT validation."""

import logging
from fastapi import Request, status, HTTPException
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials

import jwt
from mumichaspy.fastapi_jwt_chassis.config import config

logger = logging.getLogger(__name__)


def validate_and_decode_token(
    encoded_token: str,
    public_key: str = config.public_key,
    issuer: str = config.jwt_issuer,
    algorithms: str = None,
) -> dict:
    """Decode JWT if validates."""
    if algorithms is None or len(algorithms) == 0:
        algorithms = [config.jwt_algorithm]

    decoded_token = {}
    try:
        decoded_token = jwt.decode(
            encoded_token, public_key, issuer=issuer, algorithms=algorithms
        )
    except Exception as exc:
        raise_and_log_error(
            logger,
            status.HTTP_401_UNAUTHORIZED,
            "Could not decode JWT",
            f"Could not decode JWT: {exc}",
        )
    return decoded_token


class JWTBearer(HTTPBearer):  # pylint: disable=too-few-public-methods
    """HTTPBearer implementation for JWT validation."""

    def __init__(self, auto_error: bool = True):
        super().__init__(auto_error=auto_error)

    async def __call__(self, request: Request):
        # Get credentials from header
        credentials: HTTPAuthorizationCredentials = await super().__call__(request)

        # Check credential scheme
        if not credentials.scheme == "Bearer":
            raise_and_log_error(
                logger, status.HTTP_401_UNAUTHORIZED, "Invalid authentication scheme"
            )

        # Check if token is valid
        decoded_jwt = validate_and_decode_token(
            encoded_token=credentials.credentials,
            public_key=config.public_key,
            issuer=config.jwt_issuer,
            algorithms=[config.jwt_algorithm],
        )

        # Check issuer
        if decoded_jwt["iss"] != config.jwt_issuer:
            raise_and_log_error(
                logger, status.HTTP_401_UNAUTHORIZED, "Invalid JWT issuer."
            )

        # Return decoded JWT
        return decoded_jwt


class JWTBearerAdmin(JWTBearer):  # pylint: disable=too-few-public-methods
    """HTTPBearer implementation for JWT validation."""

    def __init__(self, auto_error: bool = True):
        super().__init__(auto_error=auto_error)

    async def __call__(self, request: Request):
        decoded_jwt = await super().__call__(request)
        # Check admin role
        if "ADMIN" not in decoded_jwt["roles"]:
            raise_and_log_error(
                logger, status.HTTP_403_FORBIDDEN, "Only admins can perform this action"
            )
        return decoded_jwt


def raise_and_log_error(
    my_logger, status_code: int, message: str, message_to_log: str = None
):
    """Raises HTTPException and logs an error."""

    if message_to_log is None:
        message_to_log = message

    my_logger.error(message_to_log)
    raise HTTPException(status_code, message)
